'''
Created on Jul 16, 2014

@author: posilva
'''

import re

import os
from utils import get_timestamp, generate_pkg_name

class MAVInclude(object):
    """
    """
    
    def __init__(self, filename):
        """
        """
        
        self.filename = filename
        self.name , _ = os.path.splitext( os.path.basename( self.filename ) )
        self.package_name =generate_pkg_name(self.name)  
        self.messages=[]
        self.enums=[]

class MAVMessage(object):
    """
    """
    
    def __init__(self, name, msgid, package_name,desc=None):
        """
        """
        self.name = name
        self.package_name = package_name
        self.id = msgid
        self.description = desc
        self.msg_fields = []

    def add_field(self, field):
        """
        """
        self.msg_fields.append(field)
    
    def to_msg(self):
        """
        """
        
        msg = "# Automatically Generated in " + str(get_timestamp()) + "\n" 
        msg += "# MESSAGE: " + self.name + "\n"   
        if not self.description is None:
            msg += "# Description:" + self.description.replace("\n", "\n#") + "\n"

        msg += "uint8 ID = " + self.id + "\n"        
        msg += "uint8 sysid" + "\n"
        msg += "uint8 compid" + "\n"

        for field in self.msg_fields:
            msg += field.to_string()
        return msg
    
    def filename(self):
        """
        """
        return self.name + ".msg"



class MAVEnum(object):
    """
    """
    
    def __init__(self, name, desc=None):
        """
        """
        self.name = "E_" + name
        self.description = desc
        self.entries = []


    def add_entry(self, entry):
        """
        """
        self.entries.append(entry)
    
    def to_msg(self):
        """
        """
        
        msg = "# Automatically Generated in " + str(get_timestamp) + "\n" 
        msg += "# ENUM: " + self.name + "\n"   
        if not self.description is None:
            msg += "# Description:" + self.description.replace("\n", "\n#") + "\n"
        for entry in self.entries:
            msg += entry.to_string()
        return msg
    
    def filename(self):
        """
        """
        return self.name + ".msg"

            
class MAVEntry(object):
    """
    """
    
    def __init__(self, name, value, desc=None):
        """
        """
        self.name = name
        self.description = desc
        self.value = value
        self.field_type = "uint16" if value > 255   else "uint8"
        
    def to_string(self):
        """                                                                    
        """
        msg = ""
        if not self.description is None: 
            msg += "# " + self.description + "\n"
        msg += self.field_type + " " + self.name + " = " + str(self.value) + "\n"
        return msg


data_types2ros = {
                    'char':'char',
                    'uint8_t':'uint8',
                    'uint8_t_mavlink_version':'uint8',
                    'int8_t':'int8',
                    'uint16_t':'uint16',
                    'int16_t':'int16',
                    'uint32_t':'uint32',
                    'int32_t':'int32',
                    'uint64_t':'uint64',
                    'int64_t':'int64',
                    'float':'float32',
                    'double':'float64',
                    }
data_types2mav = {
                    'char':'char',
                    'uint8':'uint8_t',
                    'int8':'int8_t',
                    'uint16':'uint16_t',
                    'int16':'int16_t',
                    'uint32':'uint32_t',
                    'int32':'int32_t',
                    'uint64':'uint64_t',
                    'int64':'int64_t',
                    'float32':'float',
                    'float64':'double',
                    }
class MAVField(object):
    """
    """
   
                        
    def is_array(self, data_type):
        """
        """
        m = re.search(r'(.*)\[(.*)\].*', data_type, re.I)
        if m:
            t = m.group(1)
            s = m.group(2)
            return (t, s)
        else: 
            return (None, None)
        if (data_type):
            pass
        
    def get_mav_type(self, data_type):
        """
        """
        t, s = self.is_array(data_type)
        if t is None:
            return data_types2mav[data_type]
        else:
            self.is_array_type = True
            self.array_size = int(s)
            return data_types2mav[t]
    
    def get_ros_type(self, data_type):
        """
        """
        t, s = self.is_array(data_type)
        if t is None:
            return data_types2ros[data_type]
        else:
            self.is_array_type = True
            self.array_size = int(s)
            return data_types2ros[t]
        
    
    def __init__(self, name, field_type, desc=None):
        """
        """
        self.name = name
        self.description = desc
        self.is_array_type = False
        self.array_size = 0
        self.field_type = self.get_ros_type(field_type)

    def to_string(self):
        """
        """
        msg = ""                                                                                     
        if not self.description is None:
            msg += "# " + self.description 
            
        data_type = self.field_type if not self.is_array_type else self.field_type + "[" + str(self.array_size) + "]"
        msg += data_type + " " + self.name + "\n"
        return msg                 




'''
Created on Jul 17, 2014

@author: posilva
'''
import unittest

from mav2ros.tools import *
class TestMAV2ROSGenerator( unittest.TestCase ):
    def test_mavgenerator( self ):
        definitions_file = '/home/posilva/ros_workspaces/roscon14_ws/libs/mavlink-1.0.11/message_definitions/v1.0/ardupilotmega.xml'
        output_dir = '/home/posilva/ros_workspaces/roscon14_ws/src'

        generator = MAVGenerator( definitions_file, output_dir )
        generator.generate(True,True);

        self.assertEqual( len( generator.parser.enums ), 6, "Enums Parsed: " + str( len( generator.parser.enums ) ) + "Expected: 6" )
        self.assertEqual( len( generator.parser.messages ), 22, "Messages Parsed: " + str( len( generator.parser.messages ) ) + "Expected: 22" )
        self.assertEqual( len( generator.parser.includes ), 1, "Includes Parsed: " + str( len( generator.parser.includes ) ) + "Expected: 1" )


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase( TestMAV2ROSGenerator )
    unittest.TextTestRunner( verbosity = 2 ).run( suite )
# unittest.main()



'''
Created on Jul 17, 2014

@author: posilva
'''
#===============================================================================
# Constants
#===============================================================================
PKG_NAME_PLACEHOLDER="$PACKAGE_NAME$"
PKG_DEFINITIONS_NAME_PLACEHOLDER="$PACKAGE_DEFINITIONS_NAME$"
CMAKE_PKG_DEPS_PLACEHOLDER="$CMAKE_PACKAGE_DEPS$"
CMAKE_PKG_ADD_MESSAGE_PLACEHOLDER="$CMAKE_PACKAGE_ADD_MESSAGES$"
CMAKE_PKG_GENERATE_MESSAGES_PLACEHOLDER="$CMAKE_PACKAGE_GENERATE_MESSAGES$"
CMAKE_PKG_ADD_NODE_PLACEHOLDER="$CMAKE_PACKAGE_ADD_NODE$"
PKG_XML_RUN_DEPS_PLACEHOLDER="$PKG_XML_RUN_DEPS$"
PKG_XML_BUILD_DEPS_PLACEHOLDER="$PKG_XML_BUILD_DEPS$"
NODE_PUBLISHERS_DECL_PLACE_HOLDER ="$NODE_PUBLISHERS_DECL$"
NODE_SUBSCRIBERS_DECL_PLACE_HOLDER ="$NODE_SUBSCRIBERS_DECL$"
NODE_MESSAGES_CALLBACKS_PLACE_HOLDER ="$NODE_MESSAGES_CALLBACKS$"
NODE_MESSAGES_CALLBACK_FIELDS_PLACE_HOLDER ="$NODE_MESSAGES_CALLBACK_FIELDS$"
NODE_MAV_SWITCH_CASES_PLACE_HOLDER = "$NODE_MAV_SWITCH_CASES$"
NODE_MAV_SWITCH_CASE_FIELDS_PLACE_HOLDER = "$NODE_MAV_SWITCH_CASE_FIELDS$"
NODE_MESSAGE_NAME_PLACE_HOLDER = "$NODE_MESSAGE_NAME$"
NODE_MESSAGE_NAME_LOWER_PLACE_HOLDER = "$NODE_MESSAGE_NAME_LOWER$"

#===============================================================================
# CMakeLists.txt template
#===============================================================================
cmakelists_template = """cmake_minimum_required(VERSION 2.8.3)
project($PACKAGE_NAME$)

find_package(catkin REQUIRED COMPONENTS
  roscpp
  rospy
  message_generation
  $CMAKE_PACKAGE_DEPS$
)

add_message_files(
   FILES
   $CMAKE_PACKAGE_ADD_MESSAGES$
)
generate_messages(
    DEPENDENCIES
    std_msgs
    $CMAKE_PACKAGE_GENERATE_MESSAGES$
)
catkin_package(
  INCLUDE_DIRS include
  CATKIN_DEPENDS roscpp message_runtime
)

include_directories(include)
include_directories(
  ${catkin_INCLUDE_DIRS}
)
$CMAKE_PACKAGE_ADD_NODE$

install(DIRECTORY include/${PROJECT_NAME}/
    DESTINATION ${CATKIN_PACKAGE_INCLUDE_DESTINATION}
    FILES_MATCHING PATTERN "*.h"
    PATTERN ".svn" EXCLUDE
)
"""

cmake_node_template="""
    add_executable(test_$PACKAGE_NAME$ src/$PACKAGE_NAME$_node.cpp)
    target_link_libraries(test_$PACKAGE_NAME$ ${catkin_LIBRARIES})
"""
#===============================================================================
# package.xml template
#===============================================================================
package_xml_template = """<?xml version="1.0"?>
<package>
    <name>$PACKAGE_NAME$</name>
    <version>0.0.1</version>
    <description>Mavlink generated messages for package $PACKAGE_NAME$</description>
    <maintainer email="posilva@academiafa.edu.pt">Pedro Marque da Silva</maintainer>
    <license>MIT</license>
    
    <author email="posilva@academiafa.edu.pt">Pedro Marques da Silva</author>
    <build_depend>message_generation</build_depend>
    <run_depend>message_runtime</run_depend>
    <run_depend>roscpp</run_depend>
    <run_depend>rospy</run_depend>
    <buildtool_depend>catkin</buildtool_depend>
    <build_depend>roscpp</build_depend>
    <build_depend>rospy</build_depend>
    
    $PKG_XML_RUN_DEPS$
    $PKG_XML_BUILD_DEPS$
</package>
"""
#===============================================================================
# Source code template for ROS node
#===============================================================================
ros_node_template="""
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 
/**
 * 
 **/
#include <ros/ros.h>
#include <$PACKAGE_NAME$/$PACKAGE_DEFINITIONS_NAME$/mavlink.h>
#include <$PACKAGE_NAME$/mavlink2ros.h>

/**
 * 
 **/
int sockfd;
int portno=5763;
int n_bytes=0;
int buff_readed=0;
struct sockaddr_in serv_addr;
struct hostent *server;

char buffer[256];

/**
 * 
 **/
void error(const char *msg) {
    perror(msg);
    exit(0);
}

/**
 * 
 **/
char read_char()
{
    if (sockfd<0)
        error("Socket not ready");

    if (buff_readed>=n_bytes){
        buff_readed=0;
        memset(buffer,0,256);
        n_bytes = read(sockfd,buffer,255);
        if(n_bytes < 0)
            error("Failed to read from socket");
    }
    return buffer[buff_readed++];
}

/**
 * 
 **/
int write_to_mav (uint8_t * b, int sz){
    if (sockfd<0)
        error("Socket not ready");
    int rc = write(sockfd,b,sz);
    ROS_INFO("Writen to MAV %d bytes",rc);
    return rc;
}        

/**
 * 
 **/
void setup_comms(){
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    
    if (sockfd < 0) 
        error("ERROR opening socket");
    
    server = gethostbyname("localhost");
    if (server == NULL) {
        error("ERROR no such host");
    }

    memset((char *) &serv_addr,0, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    memcpy((char *)server->h_addr,(char *)&serv_addr.sin_addr.s_addr,server->h_length);
    serv_addr.sin_port = htons(portno);
    
    if (connect(sockfd,(struct sockaddr *) &serv_addr,sizeof(serv_addr)) < 0) 
        error("ERROR connecting");
}

/**
 *
 **/
void close_comms(){
    if(sockfd>=0){
        close(sockfd);
    }
}

$NODE_MESSAGES_CALLBACKS$

/**
 *
 *
 */
int main(int argc, char **argv)
{
    ros::init(argc, argv, "$PACKAGE_NAME$_node");
    ros::NodeHandle n;
    ros::Rate loop_rate(100);


    setup_comms();

    /**
     * Messages Publishers Declaration
     */
$NODE_PUBLISHERS_DECL$
    /**
     * Messages Subscribers Declaration
     */
$NODE_SUBSCRIBERS_DECL$

    mavlink_message_t msg;
    mavlink_status_t status;
            
    while (ros::ok())
    {
        
        uint8_t c = read_char();
        // Try to get a new message
        if(mavlink_parse_char(MAVLINK_COMM_0, c, &msg, &status)) {
            // Handle message
            switch(msg.msgid)
            {
                $NODE_MAV_SWITCH_CASES$
                default:
                    //Do nothing
                    break;
            }
        }
        ros::spinOnce();
        loop_rate.sleep();
    }
    close_comms();
    return 0;
}

"""
#===============================================================================
# ROS messages callback template
#===============================================================================
messages_callback_template = """
/**
 *
 */            
void to_mav_$NODE_MESSAGE_NAME_LOWER$_callback(const $PACKAGE_NAME$::$NODE_MESSAGE_NAME$::ConstPtr& msg)
{
    ROS_INFO("[$PACKAGE_NAME$] Received a  'to_mav_$NODE_MESSAGE_NAME$ request");
    
    uint8_t data[MAVLINK_MAX_PACKET_LEN];
    mavlink_message_t m;
    
    m.sysid=msg->sysid;
    m.compid=msg->compid;
    
    mavlink_$NODE_MESSAGE_NAME_LOWER$_t $NODE_MESSAGE_NAME_LOWER$_out;
    
    /** ASSIGN FIELDS **/
    
$NODE_MESSAGES_CALLBACK_FIELDS$
    
    mavlink_msg_$NODE_MESSAGE_NAME_LOWER$_encode(msg->sysid, msg->compid, &m, &$NODE_MESSAGE_NAME_LOWER$_out);
    uint16_t len=mavlink_msg_to_send_buffer(data, &m);
    write_to_mav(data,len);
    
}"""
#===============================================================================
# Switch case template to mavlink messages reception
#===============================================================================
switch_case_template = """
                    case MAVLINK_MSG_ID_$NODE_MESSAGE_NAME$:
                    {
                        $PACKAGE_NAME$::$NODE_MESSAGE_NAME$ m;
                        
                        m.sysid=msg.sysid;
                        m.compid=msg.compid;
                        
                        mavlink_$NODE_MESSAGE_NAME_LOWER$_t $NODE_MESSAGE_NAME_LOWER$_in;
                        memset(&$NODE_MESSAGE_NAME_LOWER$_in, 0, sizeof($NODE_MESSAGE_NAME_LOWER$_in));
                        mavlink_msg_$NODE_MESSAGE_NAME_LOWER$_decode(&msg, &$NODE_MESSAGE_NAME_LOWER$_in);
                        
$NODE_MAV_SWITCH_CASE_FIELDS$
                        
                        from_mav_$NODE_MESSAGE_NAME_LOWER$_pub.publish(m);     
                    }
                    break;"""
#===============================================================================
# ROS Publisher declaration template
#===============================================================================
publisher_decl_template ="""\tros::Publisher from_mav_$NODE_MESSAGE_NAME_LOWER$_pub = n.advertise<$PACKAGE_NAME$::$NODE_MESSAGE_NAME$>("/from_mav_$NODE_MESSAGE_NAME_LOWER$", 10);
"""

#===============================================================================
# ROS Subscriber declaration template
#===============================================================================
subscriber_decl_template ="""\tros::Subscriber to_mav_$NODE_MESSAGE_NAME_LOWER$_sub = n.subscribe("/to_mav_$NODE_MESSAGE_NAME_LOWER$", 10, to_mav_$NODE_MESSAGE_NAME_LOWER$_callback);
"""


'''
Created on Jul 17, 2014

@author: posilva
'''

import xml.etree.ElementTree as ET
import os
from model import MAVEntry, MAVEnum, MAVMessage, MAVField, MAVInclude, data_types2mav

from utils import generate_file,get_timestamp,generate_pkg_name, mk_dirs
from templates import *



class MavlinkGenerator( object ):
    """
    """
    def __init__( self, definitions,include_dir, version = "1.0.11" ):
        self.definitions = definitions
        self.include_dir = include_dir
        self.version = version
        pass
    def generate( self ):
        url = "https://github.com/mavlink/mavlink/archive/" + self.version + ".tar.gz"
        mavlink_file = os.path.basename( url );

        import tempfile
        temp_dir = tempfile.mkdtemp()
        mavlink_download_file = temp_dir + os.sep + mavlink_file

        try:
            import urllib2
            mavlink_targz = urllib2.urlopen( url )
            output = open( mavlink_download_file, 'wb' )
            output.write( mavlink_targz.read() )
            output.close()

            if not os.path.exists( mavlink_download_file ):
                raise "Failed download file: '" + mavlink_download_file + "'"
        except :
            raise "Failed download file: '" + mavlink_download_file + "' from '" + url + "'"

        import tarfile
        tar = tarfile.open( mavlink_download_file, 'r|gz' )
        tar.extractall( temp_dir )
        root_dir = temp_dir + os.sep + tar.members[0].name

        mav_generator = root_dir + os.sep + "generator" + os.sep + "mavgen.py"
        if ( not os.path.exists( mav_generator ) ):
            raise "Generator not found: '" + mav_generator + "'"

        gen_dir = self.include_dir
        cmd = "" + mav_generator + "  --lang=C --wire-protocol=1.0 --output=" + gen_dir + " " + self.definitions

        import subprocess
        proc = subprocess.Popen( args = cmd, stdout = subprocess.PIPE, shell = True )
        ( out, _ ) = proc.communicate()
        return out

class ROSMessagesGenerator( object ):
    """
    """
    def __init__( self, messages, pkg_dir ):
        self.messages = messages
        self.pkg_dir = pkg_dir
        pass

    def generate( self ):
        msg_dir = self.pkg_dir + os.sep + "msg"
        mk_dirs( msg_dir )
        for msg in self.messages:
            filename = msg_dir + os.sep + msg.name + ".msg"
            generate_file( filename, msg.to_msg() )

class ROSPackageGenerator( object ):
    """
    """
    def __init__( self, mav_generator ):
        self.mav_generator = mav_generator

        pass


    def __gen_cmake_add_messages( self ):
        add_messages = ""
        for message in self.mav_generator.parser.messages + self.mav_generator.parser.enums:
            add_messages += "\t" + message.name + ".msg" + os.linesep
        return add_messages


    def __gen_cmake_node( self, include_node = False ):
        cmake_node = ""
        if ( include_node ):
            cmake_node = cmake_node_template
        return cmake_node

    def generate( self, include_node = False ):

        messages_deps = ""
        run_deps = ""
        build_deps = ""
        for include in self.mav_generator.parser.includes:
            messages_deps += include.package_name + os.linesep
            build_deps += "\t" + "<build_depend>" + include.package_name + "</build_depend>" + os.linesep
            run_deps += "\t" + "<run_depend>" + include.package_name + "</run_depend>" + os.linesep

        cmakelists = cmakelists_template.replace( CMAKE_PKG_ADD_MESSAGE_PLACEHOLDER, self.__gen_cmake_add_messages() ).\
                                  replace( CMAKE_PKG_ADD_NODE_PLACEHOLDER, self.__gen_cmake_node( include_node ) ).\
                                  replace( CMAKE_PKG_GENERATE_MESSAGES_PLACEHOLDER, messages_deps ).\
                                  replace( CMAKE_PKG_DEPS_PLACEHOLDER, messages_deps ).\
                                  replace( PKG_NAME_PLACEHOLDER, self.mav_generator.package_name )

        generate_file( self.mav_generator.package_dir + os.sep + "CMakeLists.txt", cmakelists )

        package_xml = package_xml_template.replace( PKG_XML_RUN_DEPS_PLACEHOLDER, run_deps ).\
                                  replace( PKG_XML_BUILD_DEPS_PLACEHOLDER, build_deps ).\
                                  replace( PKG_NAME_PLACEHOLDER, self.mav_generator.package_name )

        generate_file( self.mav_generator.package_dir + os.sep + "package.xml", package_xml )


class ROSAPIGenerator( object ):
    """
    """

    def __init__( self, mav_generator ):
        self.mav_generator = mav_generator

    def __gen_header_file( self, header_name ):
        """
        """

        header_file = "// Automatically Generated in " + str( get_timestamp() ) + os.linesep
        include_guard = self.mav_generator.package_name.upper() + "_" + header_name.upper() + "_H"
        header_file += "#ifndef " + include_guard + os.linesep
        header_file += "#define " + include_guard + os.linesep

        for include in self.mav_generator.parser.includes:
            for message in ( include.messages + include.enums ):
                header_file += "#include <" + include.package_name + os.sep + message.name + ".h>" + os.linesep

        for message in ( self.mav_generator.parser.messages + self.mav_generator.parser.enums ):
            header_file += "#include <" + self.mav_generator.package_name + os.sep + message.name + ".h>" + os.linesep

        header_file += "#endif // " + include_guard + os.linesep
        return header_file
    
    def __gen_publishers_decl( self, message_list ):
        publishers=""
        for message in message_list:
            publishers+=publisher_decl_template.replace( NODE_MESSAGE_NAME_LOWER_PLACE_HOLDER, message.name.lower() ).\
                                                replace( NODE_MESSAGE_NAME_PLACE_HOLDER, message.name ).\
                                                replace( PKG_NAME_PLACEHOLDER, message.package_name )

        return publishers
        
    
    def __gen_subscribers_decl( self, message_list ):
        subscribers =""
        for message in message_list:
            subscribers+=subscriber_decl_template.replace( NODE_MESSAGE_NAME_LOWER_PLACE_HOLDER, message.name.lower() ).\
                                                replace( NODE_MESSAGE_NAME_PLACE_HOLDER, message.name ).\
                                                replace( PKG_NAME_PLACEHOLDER, message.package_name )
        return subscribers
        
        
     
    def __gen_source_file( self ):
        message_list=[]
        
        for include in self.mav_generator.parser.includes:
            message_list+=include.messages
        message_list+=self.mav_generator.parser.messages
        
        source_file = ros_node_template.\
                        replace( NODE_MESSAGES_CALLBACKS_PLACE_HOLDER, self.__gen_callbacks(message_list) ).\
                        replace( NODE_PUBLISHERS_DECL_PLACE_HOLDER, self.__gen_publishers_decl(message_list) ).\
                        replace( NODE_SUBSCRIBERS_DECL_PLACE_HOLDER, self.__gen_subscribers_decl(message_list) ).\
                        replace( NODE_MAV_SWITCH_CASES_PLACE_HOLDER, self.__gen_switch_cases(message_list) ).\
                        replace( PKG_NAME_PLACEHOLDER, self.mav_generator.package_name ).\
                        replace( PKG_DEFINITIONS_NAME_PLACEHOLDER, self.mav_generator.parser.definition_name )
        return source_file

    def __gen_switch_case_fields( self, message ):
        switch_case_fields = ""
        for field in message.msg_fields:
            if (not field.is_array_type):
                switch_case_fields +="\tm."+field.name +" = $NODE_MESSAGE_NAME_LOWER$_in."+field.name +";\n"
            else:
                switch_case_fields +="\tmemcpy(&(m."+field.name +"), &($NODE_MESSAGE_NAME_LOWER$_in."+field.name +"), sizeof("+data_types2mav[field.field_type]+")*"+str(field.array_size)+");\n"

        return switch_case_fields

    def __gen_switch_cases( self, message_list ):
        switch_cases = ""
        for message in message_list:
            case = switch_case_template.\
            replace( NODE_MAV_SWITCH_CASE_FIELDS_PLACE_HOLDER, self.__gen_switch_case_fields( message ) ).\
            replace( NODE_MESSAGE_NAME_LOWER_PLACE_HOLDER, message.name.lower() ).\
            replace( NODE_MESSAGE_NAME_PLACE_HOLDER, message.name ).\
            replace( PKG_NAME_PLACEHOLDER, message  .package_name )
            switch_cases += case

        return switch_cases

    def __gen_callback_message_fields( self, message ):
        message_fields = ""
        for field in message.msg_fields:
            if (not field.is_array_type):
                message_fields +="\t$NODE_MESSAGE_NAME_LOWER$_out."+field.name +" = msg->"+field.name +";\n"
            else:
                message_fields +="\tmemcpy(&($NODE_MESSAGE_NAME_LOWER$_out."+field.name +"), &(msg->"+field.name +"[0]), sizeof("+data_types2mav[field.field_type]+")* (int)(msg->"+field.name +".size()));\n"
        
        return message_fields

    def __gen_callbacks( self, message_list ):
        messages_callback = ""
        for message in message_list:
            callback = messages_callback_template.replace( NODE_MESSAGES_CALLBACK_FIELDS_PLACE_HOLDER, self.__gen_callback_message_fields( message ) ).\
                                                  replace( NODE_MESSAGE_NAME_LOWER_PLACE_HOLDER, message.name.lower() ).\
                                                  replace( NODE_MESSAGE_NAME_PLACE_HOLDER, message.name ).\
                                                  replace( PKG_NAME_PLACEHOLDER, message.package_name )
            messages_callback += callback

        return messages_callback

    def generate( self ):

        header_name = "mavlink2ros"
        generate_file( self.mav_generator.include_dir + os.sep + header_name + ".h" , self.__gen_header_file( header_name ) )
        
        generate_file( self.mav_generator.source_dir + os.sep + self.mav_generator.package_name + "_node.cpp" , self.__gen_source_file() )

        pass

class MAVParser( object ):
    """
    """
    def __init__( self, package_name, definitions = None ):
        """
        """
        self.definitions = None
        self.package_name = package_name
        self.set_definitions( definitions )
        self.messages = [];
        self.enums = [];
        self.includes = [];
        # get definitions root node

    def set_definitions( self, definitions ):
        """
        """
        if ( not definitions is None ):
            self.definitions = definitions
            self.definition_name, _ = os.path.splitext( os.path.basename( self.definitions ) )
            self.definitions_dir = os.path.dirname( os.path.abspath( self.definitions ) )
            tree = ET.parse( self.definitions )
            self.mavlink_node = tree.getroot()
            if ( self.mavlink_node.tag != "mavlink" ):
                raise "Invalid root tag for a mavlink definitions file: Expected 'mavlink'"



    def __internal_parse( self ):
        for include in self.mavlink_node.iter( 'include' ):
            include_file = self.definitions_dir + os.sep + include.text;
            mav_include = MAVInclude( include_file )
            self.includes.append( mav_include )
        for node in self.mavlink_node.iter( 'enum' ):
            self.__parse_message( node )
        for node in self.mavlink_node.iter( 'message' ):
            self.__parse_message( node )

        pass

    def parse( self, definitions = None ):
        """
        """

        if ( definitions is None and self.definitions is None ):
            raise "No valid file given to parse"
        else:
            if ( self.definitions is None ):
                self.set_definitions( definitions )
            else:
                self.set_definitions( self.definitions )

            self.__internal_parse()
        return True


    def __parse_message( self, node ):
        """
        """
        is_message = ( node.tag == 'message' )
        is_enum = ( node.tag == 'enum' )

        if not ( is_enum or is_message ) :
            raise "Invalid node must have 'message' or 'enum' tag"

        if 'name' not in node.attrib:
            raise "Missing required attribute: 'name'"

        desc = node.find( 'description' )
        if not desc is None:
            desc = desc.text

        if is_message:
            mav_message = MAVMessage( node.attrib['name'] , node.attrib['id'], self.package_name, desc )
            fields = node.findall( "./field" )
            if len( fields ) > 0 :
                for field in fields:
                    desc = field.find( 'description' )
                    if not desc is None:
                        desc = desc.text
                    mav_field = MAVField( field.attrib['name'], field.attrib['type'], desc )
                    mav_message.add_field( mav_field )
            self.messages.append( mav_message )
        elif is_enum:
            mav_enum = MAVEnum( node.attrib['name'], desc )
            entries = node.findall( "./entry" )
            if len( entries ) > 0 :
                default_entry_value = 0
                for entry in entries:
                    desc = entry.find( 'description' )
                    if not desc is None:
                        desc = desc.text
                    if 'value' not in entry.attrib :
                        default_entry_value += 1
                        entry_value = default_entry_value;
                    else:
                        entry_value = int( entry.attrib['value'] )
                        default_entry_value = int( entry_value )

                    mav_entry = MAVEntry( node.tag.upper()[0] + "_" + entry.attrib['name'], entry_value, desc )
                    mav_enum.add_entry( mav_entry )
            self.enums.append( mav_enum )

class MAVGenerator( object ):
    """
    """

    def __init__( self, definitions , output_dir ):
        """
        """
        if ( definitions is None ):
            raise "Invalid definitions file: cannot be NoneType"

        if ( output_dir is None ):
            raise "Invalid Output dir: cannot be NoneType"

        self.output_dir = output_dir

        filename, _ = os.path.splitext( os.path.basename( definitions ) )
        # setup ROS packege name
        self.package_name = generate_pkg_name( filename )
        # setup ROS package paths
        self.package_dir = output_dir + os.sep + self.package_name
        self.include_dir = self.package_dir + os.sep + "include" + os.sep + self.package_name
        self.source_dir = self.package_dir + os.sep + "src"
        # force path creation
        mk_dirs( self.package_dir )
        mk_dirs( self.include_dir )
        mk_dirs( self.source_dir )
        # create parser instance
        self.parser = MAVParser( self.package_name, definitions )

    def generate( self, with_node = False , with_mavlink=False):
        """
        """
        if not self.parser.parse():
            raise "File parsing failed, reason: " + self.parser.definitions

        for include in self.parser.includes:
            include_generator = MAVGenerator( include.filename, self.output_dir )
            include_generator.generate();

            include.messages = include_generator.parser.messages
            include.enums = include_generator.parser.enums

        msg_generator = ROSMessagesGenerator( self.parser.messages + self.parser.enums , self.package_dir );
        msg_generator.generate();

        package_generator = ROSPackageGenerator( self )
        package_generator.generate(with_node)

        if ( with_node ):
            api_generator = ROSAPIGenerator( self )
            api_generator.generate()
            
        if (with_mavlink):
            mavlink_generator  = MavlinkGenerator(self.parser.definitions,self.include_dir)
            mavlink_generator.generate()



'''
Created on Jul 17, 2014

@author: posilva
'''
from distutils.core import setup

setup(
    name='mav2rosgenerator',
    version='0.1.0',
    author='Pedro Marques da Silva',
    author_email='posilva@gmail.com',
    packages=['mav2ros', 'mav2ros.test'],
    scripts=[],
    url='http://pypi.python.org/pypi/TowelStuff/',
    license='LICENSE.txt',
    description='A Mavlink protocol to ROS API Generator.',
    long_description=open('README.txt').read(),
    install_requires=[],
)